"""
Boxfile writing methods
Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
"""
#! /usr/bin/env python
#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
import csv
import os
import typing

import numpy as np
from typing import List, Union
import cryolo.utils as utils
from cryolo.utils import Filament, BoundBox
from pyStarDB import sp_pystardb as star
import pandas as pd
from abc import ABC, abstractmethod

class BoxPosition(ABC):
    @abstractmethod
    def get_center_coords(self):
        pass

    @abstractmethod
    def get_extent(self):
        pass

def write_box_yolo(path, boxes, write_star=False):
    """
    Write box/star files.
    :param path: Filepath or filename of the box file to write
    :param boxes: Boxes to write
    :param outpath: When path is a filename, it one can specifiy the output path with outpath.
    :param write_star: If true, a star file is written.
    :return: None
    """
    if write_star:
        path = path[:-3] + "star"
        write_star_file(path, boxes)

    else:
        write_eman1_boxfile(path, boxes)


def get_star_file_header(file_name):
    """
        Load the header information.
        Arguments:
        file_name - Path to the file that contains the header.
        Returns:
        List of header names, rows that are occupied by the header.
    """
    start_header = False
    header_names = []
    idx = None

    with open(file_name, "r") as read:
        for idx, line in enumerate(read.readlines()):
            if line.startswith("_"):
                if start_header:
                    header_names.append(line.strip().split()[0])
                else:
                    start_header = True
                    header_names.append(line.strip().split()[0])
            elif start_header:
                break

    if not start_header:
        raise IOError(f"No header information found in {file_name}")

    return header_names, idx

def write_star_file(path, boxes):
    is_3d =  boxes and boxes[0].z is not None
    columns=['_rlnCoordinateX', '_rlnCoordinateY']
    if is_3d is True:
        columns = ['_rlnCoordinateX', '_rlnCoordinateY','_rlnCoordinateZ']

    coords = np.zeros(shape=(len(boxes),len(columns)))
    for box_index, box in enumerate(boxes):
        coords[box_index, 0] = box.x + box.w / 2
        coords[box_index, 1] = box.y + box.h / 2

    if is_3d:
        for box_index, box in enumerate(boxes):
            coords[box_index, 2] = box.z

    df = pd.DataFrame(coords, columns=columns)
    sfile = star.StarFile(path)
    sfile.update('', df, True)
    sfile.write_star_file(overwrite=True)

def read_star_file(path, box_size=200):
    """
    Read a box file from STAR file
    :param path: Path to star file
    :param box_size
    """
    sfile = star.StarFile(path)
    panda_dataframe=sfile['']                   # the panda dataframe has noname (i.e.:data_)
    if len(panda_dataframe.iloc[:]) == 0:
        return []

    boxes=[]

    # 200 is the default value of the box size
    box_width = box_size
    box_height = box_size
    is_3d = '_rlnCoordinateZ' in panda_dataframe.columns
    for i in range(len(panda_dataframe.iloc[:])):
        b=BoundBox(x=panda_dataframe.iloc[i]['_rlnCoordinateX'] - box_width / 2, y=panda_dataframe.iloc[i]['_rlnCoordinateY']  - box_height / 2, w=box_width, h=box_height)
        if is_3d:
            b.z = panda_dataframe.iloc[i]['_rlnCoordinateZ'] # the constructor of BoundBox has not 'z' parameter
        boxes.append(b)

    return boxes

def is_star_filament_file(path):
    try:
        star.StarFile(path)
    except:
        return False

    if path.endswith(".star") == False:
        return False
    return True

def read_star_filament_file_(path, box_width, min_distance=0):
    sfile = star.StarFile(path)

    xcoords = sfile['']["_rlnCoordinateX"]
    ycoords = sfile['']["_rlnCoordinateY"]
    filaments = []
    for i in range(len(xcoords)):
        if i % 2 == 0:
            x_start = int(xcoords[i] - box_width / 2)
            y_start = int(ycoords[i] - box_width / 2)
            box_start = BoundBox(x=x_start, y=y_start, w=box_width, h=box_width)
        else:
            x_end = int(xcoords[i] - box_width / 2)
            y_end = int(ycoords[i] - box_width / 2)
            box_end = BoundBox(x=x_end, y=y_end, w=box_width, h=box_width)
            length = np.sqrt((x_start - x_end) ** 2 + (y_start - y_end) ** 2)
            if min_distance == 0:
                parts = int((length / (0.1 * box_width)))
            else:
                parts = length / min_distance
            boxes = utils.getEquidistantBoxes(
                box1=box_start, box2=box_end, num_boxes=parts
            )
            f = Filament(boxes)
            filaments.append(f)
    return filaments

def write_star_filemant_file_(path, filaments):
    num_boxes = 2*len(filaments)

    coords = np.zeros(shape=(num_boxes,2))
    i = 0
    for fil in filaments:
        bstart = fil.boxes[0]
        bend = fil.boxes[-1]
        coords[i, 0] = bstart.x + bstart.w / 2
        coords[i, 1] = bstart.y + bstart.h / 2
        coords[i+1, 0] = bend.x + bend.w / 2
        coords[i+1, 1] = bend.y + bend.h / 2
        i = i + 2

    df = pd.DataFrame(coords, columns=['_rlnCoordinateX', '_rlnCoordinateY'])
    sfile = star.StarFile(path)
    sfile.update('', df, True)
    sfile.write_star_file(overwrite=True)


def read_eman1_boxfile(path, is_SPA=True, box_size_default=200) -> List[BoundBox]:
    """
    Read a box file in EMAN1 box format.
    :param path: Path to box file
    :param is_SPA: False if the file is referred to a tomo
    :param box_size_default: boxsize for tomo box files
    :return: List of bounding boxes
    """
    boxreader = np.atleast_2d(np.genfromtxt(path))
    boxes =[]
    len_row = len(boxreader[0])
    if is_SPA: #2D case is always with width and height values
        boxes = [BoundBox(x=box[0], y=box[1], w=box[2], h=box[3]) for box in boxreader]
    elif len_row in [3,5]: #3D case without width and height values
        w= box_size_default if len_row == 3 else boxreader[0][3]
        h = box_size_default if len_row == 3 else boxreader[0][4]
        boxes = [BoundBox(x=box[0]-w/2, y=box[1]-h/2, z=box[2], w=w, h=h) for box in boxreader]
    return boxes

def _create_cbox_boundbox(box):
    bound_box = BoundBox(x=box[0], y=box[1], w=box[2], h=box[3], c=box[4])
    if len(box) > 5:
        bound_box.meta["est_box_size"] = (box[5], box[6])
    return bound_box

def read_cbox_include_list(path):
    starfile = star.StarFile(path)
    return starfile['cryolo_include']["_slice_index"].tolist()

def read_star_filament_file(path, box_width, min_distance=0) -> typing.List[Filament]:
    sfile = star.StarFile(path)
    filaments = {}
    for i in range(len(sfile['']['_rlnCoordinateX'])):
        c = 1
        if '_rlnAutopickFigureOfMerit' in sfile['']:
            c= sfile['']['_rlnAutopickFigureOfMerit'][i]
        box = BoundBox(
            x=sfile['']['_rlnCoordinateX'][i]-box_width/2,
            y=sfile['']['_rlnCoordinateY'][i]-box_width/2,
            w=box_width,
            h=box_width,
            c=c
        )


        fid = int(sfile['']['_rlnHelicalTubeID'][i])
        if fid in filaments:
            filaments[fid].append(box)
        else:
            filaments[fid] = [box]
    filaments_list = [Filament(filaments[key]) for key in filaments]
    if min_distance > 0:
        filaments_list = utils.resample_filaments(filaments_list, min_distance)

    return filaments_list


def write_star_filemant_file(path, filaments):
    columns = ['_rlnCoordinateX', '_rlnCoordinateY', '_rlnAutopickFigureOfMerit', '_rlnHelicalTubeID', '_rlnAnglePsiPrior']
    data_dict = {}
    for clm in columns:
        data_dict[clm] = []

    for fil_id, fil in enumerate(filaments):
        utils.set_filament_psi_priors(fil)
        for box_index, box in enumerate(fil.boxes):
            data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
            data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
            data_dict['_rlnAutopickFigureOfMerit'].append(box.c)
            data_dict['_rlnHelicalTubeID'].append(fil_id + 1)
            data_dict['_rlnAnglePsiPrior'].append(fil.meta['psi'][box_index])

    df = pd.DataFrame(data_dict)
    sfile = star.StarFile(path)
    sfile.update('', df, True)
    sfile.write_star_file(overwrite=True)


def write_cryosparc_star_file(path: str, coordinates: Union[List[BoundBox], List[Filament]], micrograph_name: str):
    is_filament_data = isinstance(coordinates[0], Filament)
    columns = ['_rlnMicrographName', '_rlnCoordinateX', '_rlnCoordinateY', '_rlnAutopickFigureOfMerit']
    if is_filament_data:
        columns.append('_rlnHelicalTubeID')
        columns.append('_rlnAnglePsiPrior')
    data_dict = {}
    fid_offset = 0
    if os.path.exists(path):
        starfile = star.StarFile(path)
        if '_rlnHelicalTubeID' in starfile[''] and len(starfile['']['_rlnHelicalTubeID'])>0:
            fid_offset=np.max(starfile['']['_rlnHelicalTubeID'])

    for clm in columns:
        data_dict[clm] = []
    if is_filament_data:
        for fil_id, fil in enumerate(coordinates):
            utils.set_filament_psi_priors(fil)
            m_name = micrograph_name[fil_id]
            for box_index, box in enumerate(fil.boxes):
                data_dict['_rlnMicrographName'].append(m_name)
                data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
                data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
                data_dict['_rlnAutopickFigureOfMerit'].append(box.c)
                data_dict['_rlnHelicalTubeID'].append(fid_offset+fil_id+1)
                data_dict['_rlnAnglePsiPrior'].append(fil.meta['psi'][box_index])
    else:
        for box_index, box in enumerate(coordinates):
            data_dict['_rlnMicrographName'].append(micrograph_name[box_index])
            data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
            data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
            data_dict['_rlnAutopickFigureOfMerit'].append(box.c)

    if os.path.exists(path):
        df = starfile['']
        df = df.append(pd.DataFrame(data_dict))
    else:
        df = pd.DataFrame(data_dict)

    sfile = star.StarFile(path)
    sfile.update('', df, True)
    sfile.write_star_file(overwrite=True)


def read_cbox_boxfile(path, min_dist_boxes_filament=0) -> Union[List[BoundBox], List[Filament]]:
    """

    :param path: Path to read
    :param min_dist_boxes_filament: Minimum distance between filament boxes (in case it is filament)
    :return: Either a list of bounding boxes or a list of filaments.
    """
    args_box_mapping = {
        '_CoordinateX': "x",
        '_CoordinateY': "y",
        '_CoordinateZ': 'z',
        '_Width': 'w',
        '_Height': 'h',
        '_Confidence': 'c',
        '_Depth': 'depth',
    }
    args_meta_mapping = {
        '_EstWidth': 'est_box_size',
        '_EstHeight': 'est_box_size',
        '_NumBoxes': 'num_boxes',
        '_Angle': "angle"
    }
    try:
        starfile = star.StarFile(path)
        boxes = []
        filaments = {}
        if 'cryolo' in starfile:
            is_filament = '_filamentid' in starfile['cryolo']
            for i in range(len(starfile['cryolo']['_CoordinateX'])):

                kwargs = {}
                for arg in args_box_mapping:
                    if arg in starfile['cryolo']:
                        if starfile['cryolo'][arg][i] is None or np.isnan(starfile['cryolo'][arg][i]):
                            continue
                        kwargs[args_box_mapping[arg]] = starfile['cryolo'][arg][i]

                box = BoundBox(**kwargs)

                for arg in args_meta_mapping:
                    if arg in starfile['cryolo']:

                        if starfile['cryolo'][arg][i] is None or np.isnan(starfile['cryolo'][arg][i]):
                            continue
                        if args_meta_mapping[arg] in box.meta:
                            try:
                                box.meta[args_meta_mapping[arg]].extend(starfile['cryolo'][arg][i])
                            except Exception:
                                l = [box.meta[args_meta_mapping[arg]]]
                                l.append(starfile['cryolo'][arg][i])
                                box.meta[args_meta_mapping[arg]] = l
                        else:

                            box.meta[args_meta_mapping[arg]] = starfile['cryolo'][arg][i]

                if is_filament:
                    fid = int(starfile['cryolo']['_filamentid'][i])
                    if fid in filaments:
                        filaments[fid].append(box)
                    else:
                        filaments[fid] = [box]
                else:
                    boxes.append(box)

            if is_filament:
                filaments_list = [Filament(filaments[key]) for key in filaments]
                if min_dist_boxes_filament > 0:
                    filaments_list = utils.resample_filaments(filaments_list, min_dist_boxes_filament)

                return filaments_list
            else:
                return boxes
    except Exception :
        print("Reading old CBOX format file")
        return read_cbox_boxfile_old(path)


def read_cbox_boxfile_old(path):
    """
    Read a box file in EMAN1 box format.
    :param path: Path to box file
    :return: List of bounding boxes
    """
    boxreader = np.atleast_2d(np.genfromtxt(path))

    boxes = [_create_cbox_boundbox(box) for box in boxreader]
    return boxes


def write_eman1_boxfile(path, boxes):
    with open(path, "w") as boxfile:
        boxwriter = csv.writer(
            boxfile, delimiter="\t", quotechar="|", quoting=csv.QUOTE_NONE
        )

        for box in boxes:
            # box.x / box.y = Lower left corner
            boxwriter.writerow([box.x, box.y, box.w, box.h])

def write_eman_boxfile3d(path,boxes):

    with open(path, "w") as boxfile:

        boxwriter = csv.writer(
            boxfile, delimiter="\t", quotechar="|", quoting=csv.QUOTE_NONE
        )

        for box in boxes:
            # box.x / box.y / box.z need to be center coordinates. z is already centered after tracing
            boxwriter.writerow([box.x+box.w/2, box.y+box.h/2, box.z])

def write_coords_file(path ,boxes):
    with open(path, "w") as boxfile:

        boxwriter = csv.writer(
            boxfile, delimiter=" ", quotechar="|", quoting=csv.QUOTE_NONE, dialect=csv.unix_dialect,
        )

        for box in boxes:
            # box.x / box.y / box.z need to be center coordinates. z is already centered after tracing
            boxwriter.writerow([box.x+box.w/2, box.y+box.h/2, box.z])


def box_to_np_array(boxes, filament_id=None):
    num_boxes = len(boxes)
    num_fields = 11
    if filament_id is not None:
        num_fields = num_fields + 1
    coords = np.zeros(shape=(num_boxes, num_fields))

    for i, box in enumerate(boxes):
        coords[i, 0] = box.x
        coords[i, 1] = box.y
        coords[i, 2] = box.z
        coords[i, 3] = box.w
        coords[i, 4] = box.h
        coords[i, 5] = box.depth
        if "boxsize_estimated" in box.meta:
            coords[i, 6] = box.meta["boxsize_estimated"][0]
            coords[i, 7] = box.meta["boxsize_estimated"][1]
        else:
            coords[i, 6] = None
            coords[i, 7] = None
        coords[i, 8] = box.c
        if "num_boxes" in box.meta:
            coords[i, 9] = box.meta["num_boxes"]
        else:
            coords[i, 9] = None
        if "angle" in box.meta:
            coords[i, 10] = box.meta["angle"]
        else:
            coords[i, 10] = None
        if filament_id is not None:
            coords[i, 11] = filament_id

    return coords



def write_cbox_file(path, coordinates, additional_slice_include=None, additional_slice_exclude=None):
    """
    Write a CBOX file in STAR format.
    :param path: File path
    :param coordinates: List of BoundingBoxes / Filaments
    :param additional_slice_include: Only needed for tomograms. List of indicies that should be added\
    to the cryolo_include_list.
    :param additional_slice_exclude: Only needed for tomograms. A list of slices that should be ignored.
    :param filament_id if boxes
    """
    if len(coordinates)==0:
        open(path, 'a').close()
        return
    columns = []
    columns.append('_CoordinateX')
    columns.append('_CoordinateY')
    columns.append('_CoordinateZ')
    columns.append('_Width')
    columns.append('_Height')
    columns.append('_Depth')
    columns.append('_EstWidth')
    columns.append('_EstHeight')
    columns.append('_Confidence')
    columns.append('_NumBoxes')
    columns.append('_Angle')

    if len(coordinates)>0 and isinstance(coordinates[0], Filament):
        coords = []
        for filament_id, filament in enumerate(coordinates):
            fil_box_coords = box_to_np_array(filament.boxes,filament_id=filament_id)
            coords.append(fil_box_coords)

        coords = np.concatenate(coords)
        columns.append('_filamentid')

    else:
        coords = box_to_np_array(coordinates)

    include_slices = [a for a in np.unique(coords[:,2]).tolist() if not np.isnan(a)]

    if additional_slice_include is not None:
        include_slices.extend(additional_slice_include)
        include_slices = list(set(include_slices))

    if additional_slice_exclude is not None:
        include_slices = [i for i in include_slices if i not in additional_slice_exclude]



    sfile = star.StarFile(path)




    version_df = pd.DataFrame([["1.0"]], columns=['_cbox_format_version'])
    sfile.update('global', version_df, False)

    df = pd.DataFrame(coords, columns=columns)
    sfile.update('cryolo', df, True)


    include_df = pd.DataFrame(include_slices, columns=['_slice_index'])
    sfile.update('cryolo_include', include_df, True)

    sfile.write_star_file(overwrite=True, tags=['global','cryolo','cryolo_include'])


def write_cbox_file_(path, boxes):
    with open(path, "w") as boxfile:
        boxwriter = csv.writer(
            boxfile, delimiter="\t", quotechar="|", quoting=csv.QUOTE_NONE
        )

        for box in boxes:
            est_width = box.meta["boxsize_estimated"][0]
            est_height = box.meta["boxsize_estimated"][1]
            # box.x / box.y = Lower left corner
            boxwriter.writerow(
                [box.x, box.y, box.w, box.h, box.c, est_width, est_height]
            )


def write_boxfile_manager(path, rectangles):
    """

    :param path: Filepath
    :param rectangles: Rectangles to write
    :return: None
    """
    boxes = []
    for rect in rectangles:
        x_lowerleft = int(rect.get_x())
        y_lowerleft = int(rect.get_y())
        boxize = int(rect.get_width())
        box = BoundBox(x=x_lowerleft, y=y_lowerleft, w=boxize, h=boxize)
        boxes.append(box)
    write_eman1_boxfile(path, boxes)


def write_eman1_filament_start_end(filaments, path):
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    with open(path, "w", newline='', encoding='utf-8') as boxfile:
        boxwriter = csv.writer(
            boxfile, delimiter="\t", quotechar="|", quoting=csv.QUOTE_NONE, lineterminator="\n",
        )
        for fil in filaments:
            fil_boxes = fil.boxes

            boxwriter.writerow(
                [
                    ""
                    + str(fil_boxes[0].x)
                    + " "
                    + str(fil_boxes[0].y)
                    + " "
                    + str(fil_boxes[0].w)
                    + " "
                    + str(fil_boxes[0].h)
                    + " "
                    + str(-1)
                ]
            )

            boxwriter.writerow(
                [
                    ""
                    + str(fil_boxes[len(fil_boxes) - 1].x)
                    + " "
                    + str(fil_boxes[len(fil_boxes) - 1].y)
                    + " "
                    + str(fil_boxes[len(fil_boxes) - 1].w)
                    + " "
                    + str(fil_boxes[len(fil_boxes) - 1].h)
                    + " "
                    + str(-2)
                ]
            )

def is_star_filament_file(path: str) -> bool:
    if not path.endswith(".star"):
        return False

    sfile = star.StarFile(path)

    return '_rlnHelicalTubeID' in sfile['']

def is_eman1_filament_start_end(path):
    if not path.endswith(".box"):
        return False
    try:
        box_lines = np.atleast_2d(np.genfromtxt(path))
        if len(box_lines) < 2:
            return False
        return (
            len(box_lines[0]) == 5 and box_lines[0][4] == -1 and box_lines[1][4] == -2
        )
    except ValueError:
        return False


def read_eman1_filament_start_end(path, min_distance=0):
    boxreader = np.atleast_2d(np.genfromtxt(path))
    filaments = []
    for box in boxreader:
        if int(box[4]) == -1:
            x_start = int(box[0])
            y_start = int(box[1])
            box_width = int(box[2])
            box_start = BoundBox(x=x_start, y=y_start, w=box_width, h=box_width)
        if int(box[4]) == -2:
            x_end = int(box[0])
            y_end = int(box[1])
            box_width = int(box[2])
            box_end = BoundBox(x=x_end, y=y_end, w=box_width, h=box_width)
            length = np.sqrt((x_start - x_end) ** 2 + (y_start - y_end) ** 2)
            if min_distance == 0:
                parts = int((length / (0.1 * box_width)))
            else:
                parts = length / min_distance
            boxes = utils.getEquidistantBoxes(
                box1=box_start, box2=box_end, num_boxes=parts
            )
            f = Filament(boxes)
            filaments.append(f)

    return filaments

def is_eman1_helicon(path):
    if not (path.endswith(".box") or path.endswith(".txt")):
        return False
    with open(path) as f:
        first_line = f.readline()
        f.close()
    return "#micrograph" in first_line


def read_eman1_helicon(path, min_distance=0):
    """

    :param path: Path to boxfiel in helicon formats
    :param min_distance: Two boxes in the filament will have at least this distance
    :return:
    """

    def get_first_and_last_coord(helix_line):
        if not helix_line.startswith("#helix"):
            raise ValueError("Line does not start with '#helix'")
        import re
        result = re.findall('\d+.\d*',helix_line)
        allnumbers = [float(r) for r in result]
        return allnumbers[:4]

    if os.stat(path).st_size != 0:
        split_indicis = []
        boxsize = 0
        index_first_helix = -1
        csvlines = None
        with open(path, "r") as csvfile:
            csvlines = csvfile.readlines()
            helixlines_indicies = []
            for index, row in enumerate(csvlines):
                if row.startswith("#segment"):
                    boxsize = int(float(row.split()[2]))
                elif row.startswith("#helix"):
                    boxsize = int(float(row[(row.rfind(",") + 1) :]))
                    if index_first_helix == -1:
                        index_first_helix = index
                    else:

                        split_indicis.append(
                            index - index_first_helix - (len(split_indicis) + 1)
                        )
                    helixlines_indicies.append(index)

        filaments = []
        coordinates = np.atleast_2d(np.genfromtxt(path))
        coordinates_lowleftcorner = coordinates - boxsize / 2
        coord_filaments = np.split(coordinates_lowleftcorner, split_indicis)
        #sqdistance = min_distance * min_distance

        for filament_index, filament in enumerate(coord_filaments):
            #print("FIL-----------START")
            f = Filament()
            first_and_last_coord = get_first_and_last_coord(csvlines[helixlines_indicies[filament_index]])

            bb = BoundBox(first_and_last_coord[0]-boxsize/2, first_and_last_coord[1]-boxsize/2, boxsize, boxsize)
            f.add_box(bb)

            for coords in filament:
                if len(coords)>0:
                    bb = BoundBox(coords[0], coords[1], boxsize, boxsize)
                    f.add_box(bb)

            bb = BoundBox(first_and_last_coord[2]-boxsize/2, first_and_last_coord[3]-boxsize/2, boxsize,
                               boxsize)
            f.add_box(bb)
            if min_distance>0:
                fr = utils.resample_filament(f,min_distance)
                f = fr
            
            filaments.append(f)

        return filaments
    return None


def write_eman1_helicon(filaments, path, image_filename):


    with open(path, "w", newline='', encoding='utf-8') as boxfile:
        boxwriter = csv.writer(
            boxfile, delimiter="\t", quotechar="|", quoting=csv.QUOTE_NONE, lineterminator="\n",
        )
        # micrograph: actin_cAla_1_corrfull.mrc
        # segment length: 384
        # segment width: 384

        if filaments is not None and len(filaments) > 0:

            boxsize = filaments[0].boxes[0].w

            boxwriter.writerow(["#micrograph: " + image_filename])
            boxwriter.writerow(["#segment length: " + str(int(boxsize))])
            boxwriter.writerow(["#segment width: " + str(int(boxsize))])

            for fil in filaments:
                if len(fil.boxes) > 0:
                    boxwriter.writerow(
                        [
                            "#helix: ("
                            + str(fil.boxes[0].x + boxsize / 2)
                            + ", "
                            + str(fil.boxes[0].y + boxsize / 2)
                            + "),"
                            + "("
                            + str(fil.boxes[len(fil.boxes) - 1].x + boxsize / 2)
                            + ", "
                            + str(fil.boxes[len(fil.boxes) - 1].y + boxsize / 2)
                            + "),"
                            + str(int(boxsize))
                        ]
                    )

                    for box in fil.boxes[1:-1]:
                        boxwriter.writerow(
                            [
                                ""
                                + str(box.x + boxsize / 2)
                                + " "
                                + str(box.y + boxsize / 2)
                            ]
                        )


from cryolo.augmentation.AugmentationMethod import AugmentationMethod
from scipy import ndimage

import numpy as np

class AverageBlurAugmentation(AugmentationMethod):

    def __init__(self, kernel_size : list):
        """
        :param kernel_size: Range for random kernel size
        """

        self.kernel_size = kernel_size

    def transform_image(self, image : np.array) -> np.array:
        """
        Applys average blurring with random kernel size
        :param image: Input image (numpy array)
        :return: Blurred image
        """
        rang_kernel_size = np.random.randint(self.kernel_size[0], self.kernel_size[1])
        image = ndimage.uniform_filter(
            image, size=rang_kernel_size, mode="nearest", output=image
        )
        return image

    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        return object_coords



import copy

import numpy as np

from cryolo.augmentation.AugmentationMethod import AugmentationMethod


class Rot90Augmentation(AugmentationMethod):

    def __init__(self, num_rotations : float):
        self.num_rotations = num_rotations

    def transform_image(self, image : np.array) -> np.array:

        if np.squeeze(image).shape[0] != np.squeeze(image).shape[1] and self.num_rotations % 2 != 0:
            raise ValueError("Rotational data augmentations failed. For non-square images the number of rotations needs even.")

        image = np.rot90(image, k=self.num_rotations, axes=(1, 0))
        return image

    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        object_coords_cpy = copy.deepcopy(object_coords)
        for obj in object_coords_cpy:
            hcenter = float(image_dims[1]) / 2
            wcenter = float(image_dims[0]) / 2

            for i in range(self.num_rotations):
                obj["xmin"] = obj["xmin"] - wcenter
                obj["xmax"] = obj["xmax"] - wcenter
                obj["ymin"] = obj["ymin"] - hcenter
                obj["ymax"] = obj["ymax"] - hcenter

                help = obj["xmin"]
                obj["xmin"] = -1 * obj["ymin"]
                obj["ymin"] = help

                help = obj["xmax"]
                obj["xmax"] = -1 * obj["ymax"]
                obj["ymax"] = help

                obj["xmin"] = obj["xmin"] + wcenter
                obj["xmax"] = obj["xmax"] + wcenter
                obj["ymin"] = obj["ymin"] + hcenter
                obj["ymax"] = obj["ymax"] + hcenter

                # Swap xmin and xmax
                help = obj["xmax"]
                obj["xmax"] = obj["xmin"]
                obj["xmin"] = help

                if "angle" in obj:
                    obj["angle"] = obj["angle"] + np.pi / 2
                    if obj["angle"] > np.pi:
                        obj["angle"] = obj["angle"] - np.pi

        return object_coords_cpy


import copy

import numpy as np
from scipy import ndimage

from cryolo.augmentation.AugmentationMethod import AugmentationMethod


class CustomRotationAugmentation(AugmentationMethod):

    def __init__(self, rotation_angle_rad : float):
        self.rotation_in_rad = rotation_angle_rad

    def transform_image(self, image : np.array) -> np.array:
        return ndimage.rotate(image, np.rad2deg(self.rotation_in_rad), reshape=False, mode='reflect', order=1)

    def transform_coords(self, object_coords : list, image_dims: list) -> list:
        hcenter = float(image_dims[1]) / 2
        wcenter = float(image_dims[0]) / 2

        object_coords_cpy = copy.deepcopy(object_coords)

        for obj in object_coords_cpy:
            # 1. calculate center coordinate of object
            rotational_center_x = wcenter
            rotational_center_y = hcenter
            center_x = obj["xmin"] + (obj["xmax"] - obj["xmin"]) / 2 - rotational_center_x
            center_y = obj["ymin"] + (obj["ymax"] - obj["ymin"]) / 2 - rotational_center_y

            # 2. rotate center point
            c, s = np.cos(-1*self.rotation_in_rad), np.sin(-1*self.rotation_in_rad)
            R = np.array(((c, -s), (s, c)))

            #rot_point = R @ np.array([center_x, center_y])
            rot_point =  np.dot(R, np.array([center_x, center_y])) #matmul leads to errors in docker ci unittesting
            rot_point[0] = rot_point[0] + rotational_center_x
            rot_point[1] = rot_point[1] + rotational_center_y

            # 4. calculate new xmin/max
            boxw = obj["xmax"] - obj["xmin"]
            boxh = obj["ymax"] - obj["ymin"]
            obj["xmin"] = rot_point[0] - boxw / 2
            obj["xmax"] = rot_point[0] + boxw / 2
            obj["ymin"] = rot_point[1] - boxh / 2
            obj["ymax"] = rot_point[1] + boxh / 2

            # 3. update angle by adding rotation range c
            if "angle" in obj:
                obj["angle"] = obj["angle"] - self.rotation_in_rad
                while obj["angle"] < 0:
                    obj["angle"] = obj["angle"] + np.pi
        return object_coords_cpy


from abc import ABC, abstractmethod
import numpy as np
from typing import Dict
class Metric(ABC):

    @abstractmethod
    def eval(self, boxes : np.array, boxesgt : np.array) -> Dict[str,float]:
        """
        :param boxes: Predicted data
        :param boxesgt: Ground truth data
        :return: List with evaluation result
        """
        pass


from cryolo.augmentation.AugmentationMethod import AugmentationMethod

import numpy as np

class AdditiveGaussianNoiseAugmentation(AugmentationMethod):

    def __init__(self, max_sigma_range_factor : float):
        """
        :param max_sigma_range_factor: Range for max_sigma_range. The standard deviation of the noise
        is choosen randomly depending on the standard deviation of the image.
        """
        self.max_sigma_range_factor = max_sigma_range_factor

    def transform_image(self, image : np.array) -> np.array:
        """
        Add random gaussian noise to image
        :param image: Input image
        choosen randomly depending on the standard deviation of the image.
        The choosen standard deviation for noise is between: 0 and  max_sigma_factor*6*np.std(image)
        :return: noise added image
        """
        width = 2 * 3 * np.std(image)
        max_sigma = width * self.max_sigma_range_factor
        rand_sigma = np.random.rand() * max_sigma
        noise = np.random.randn(image.shape[0], image.shape[1])

        # image = noise*rand_sigma + image

        np.multiply(noise, rand_sigma, out=noise)
        np.add(image, noise, out=image)

        if not np.issubdtype(image.dtype, np.float32):
            image = image.astype(np.float32, copy=False)

        return image

    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        return object_coords


#
# COPYRIGHT
#
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#
import multiprocessing
import os
import warnings

import numpy as np
import scipy.fftpack as fft
import json
import cryolo.imagereader as imagereader

def window(width, shape):
    """

    :param image:
    :param width: Width of the gaussian filter in fourier pixel
    :return:
    """
    center_image_y = shape[0] / 2
    center_image_x = shape[1] / 2
    x_coord_vec = np.linspace(0, shape[1], shape[1])
    y_coord_vec = np.linspace(0, shape[0], shape[0])
    x_coord_mat, y_coord_mat = np.meshgrid(x_coord_vec, y_coord_vec, sparse=True)
    sq1 = ((y_coord_mat - center_image_y) ** 2)
    sq2 = ((x_coord_mat - center_image_x) ** 2)
    sqs = sq1 + sq2

    sq = np.sqrt(
            sqs
        )
    dist_mat = np.float16(
        sq
        + width / 2
    )
    gmask = np.sin(np.pi * dist_mat / (width - 1))
    gmask[dist_mat > width] = 0
    return gmask


def next_power_of2(number):
    return int(np.power(2, np.ceil(np.log2(number))))


def zero_pad(img, size_new):
    pad_extends = []
    dif_shape_y = size_new[0] - img.shape[0]
    dif_shape_x = size_new[1] - img.shape[1]

    pad_extends.append((dif_shape_y // 2, dif_shape_y // 2 + dif_shape_y % 2))
    pad_extends.append((dif_shape_x // 2, dif_shape_x // 2 + dif_shape_x % 2))

    padded = np.pad(img, pad_extends, "symmetric")

    return padded, pad_extends


def apply_fft_mask(img, mask):
    """
    Filter your image with a mask
    :param img: Image to be filtered (numpy array)
    :param mask: Mask you want to apply in fft space
    :return: Return filtered image
    """
    from time import perf_counter
    t1_start = perf_counter()
    input_image_shape = img.shape

    # Zero Pad
    new_image_size = mask.shape
    padded, pad_extends = zero_pad(img, new_image_size)

    # The input image is not needed anymore at this point
    del img

    # FFT
    img_fft = fft.fft2(padded)

    # Padded image is not needed anymore
    del padded

    img_fft = fft.fftshift(img_fft)
    # Windowing with cosine window
    img_fft_windowed = img_fft * mask
    del img_fft

    # Backward fft
    img_fft_windowed = fft.ifftshift(img_fft_windowed)
    img_filtered = fft.ifft2(img_fft_windowed)

    del img_fft_windowed
    # Crop
    img_cropped = img_filtered[
        pad_extends[0][0] : (pad_extends[0][0] + input_image_shape[0]),
        pad_extends[1][0] : (pad_extends[1][0] + input_image_shape[1]),
    ]

    del img_filtered

    t1_stop = perf_counter()

    #print("Elapsed time:", t1_stop-t1_start)

    return np.real(img_cropped)

from typing import List, Dict
def validate_and_dump_state(paths: List[str], current_state: Dict):

    for file in [file for file in paths if os.path.exists(file)]:
        json_name = os.path.splitext(os.path.basename(file))[0]+".json"
        json_path = os.path.join(os.path.dirname(file), json_name)
        delete = False
        if os.path.exists(json_path):

            with open(json_path) as f:
                json_content = json.load(f)
                if json_content['resize_to_shape'] != current_state['resize_to_shape'] or json_content['filter_cutoff'] != current_state['filter_cutoff']:
                    delete=True

        else:
            delete=True

        if delete:
            os.remove(file)
            if os.path.exists(json_path):
                os.remove(json_path)

        # Dump state
        with open(json_path, 'w') as fp:
            json.dump(current_state, fp)


def filter_images(images, filter_cutoff, filter_tmp_path, num_cpu=-1, resize_to_shape=None):
    """
    Function will filter images gives in the image dictonary, write them into the filter_output_path

    :param images: list of images
    :param filter_cutoff: Absolute cutoff frequence (0-0.5)
    :param filter_tmp_path: Output directory for temporary files
    :return: None
    """

    if not os.path.isdir(filter_tmp_path):
        os.makedirs(filter_tmp_path)

    # Correct absolute filenames
    adapted_paths = [
        img["filename"][1:] if img["filename"][0] == "/" else img["filename"]
        for img in images
    ]

    # Create file paths for filtered files
    filtered_file_paths = [
        os.path.join(filter_tmp_path, img_pth) for img_pth in adapted_paths
    ]

    # Remove files with different state, create new state
    current_state = {
        "resize_to_shape": resize_to_shape,
        "filter_cutoff": filter_cutoff
    }
    validate_and_dump_state(filtered_file_paths, current_state)

    # Get index of images that are already filtered
    index_images_already_filtered = [i for i,file in enumerate(filtered_file_paths) if os.path.exists(file)]

    # Updates paths for already filtered images
    for index in index_images_already_filtered:
        images[index]["filename"] = filtered_file_paths[index]

    # Update paths that are double
    unique_paths = []
    unique_indices = []

    for image_index, path in enumerate(filtered_file_paths):
        if path not in unique_paths:
            unique_paths.append(path)
            unique_indices.append(image_index)
        else:
            images[image_index]["filename"] = filtered_file_paths[image_index]

    # Get index of images that have to be filtered
    index_images_to_filter = [
        i
        for i, pth in enumerate(filtered_file_paths)
        if not os.path.exists(pth) and
           os.path.basename(pth)[0] != "." and i in unique_indices
    ]


    # Create argument tubles for them
    arg_tubles = []
    arg_tubles.extend(
        [
            (images[i]["filename"], filter_cutoff, filter_tmp_path, -1, -1, resize_to_shape)
            for i in index_images_to_filter
        ]
    )

    # Do the filtering
    num_processes = None
    if num_cpu > 0:
        num_processes = num_cpu

    pool = multiprocessing.Pool(
        maxtasksperchild=1, processes=num_processes
    )
    all_filtered_images_paths = pool.starmap(
        filter_single_image_and_write_to_disk, arg_tubles, chunksize=1
    )
    pool.close()
    pool.join()

    # Update paths of newly filtered images
    for i, index in enumerate(index_images_to_filter):
        images[index]["filename"] = all_filtered_images_paths[i]


def filter_single_image_and_write_to_disk(
    img_path,
    filter_cutoff,
    output_path_filtered_images,
    img_num=-1,
    total_num=-1,
    resize_to_shape=None,
):
    """
    Filteres the image in img_path, save it to disk and returns the path to the filtered images
    :param img_path: Path to image to filter
    :param filter_cutoff: Absolute frequence (0 - 0.5)
    :param output_path_filtered_images: Output path for filtered images
    :param resize_to_shape: tuple with image size. the image will be downsized to this size before filtering
    :return: Path to the filtered image
    """
    dirname = os.path.dirname(img_path)
    if dirname.startswith("/"):
        dirname = dirname[1:]
    filtered_image_folder = os.path.join(output_path_filtered_images, dirname)
    filtered_img_tmp_file_path = os.path.join(
        filtered_image_folder, os.path.basename(img_path)
    )

    current_state = {
        "resize_to_shape": resize_to_shape,
        "filter_cutoff": filter_cutoff
    }
    validate_and_dump_state([filtered_img_tmp_file_path], current_state)

    if not os.path.exists(filtered_img_tmp_file_path):
        if filter_cutoff > 0.5:
            warnings.simplefilter("once", UserWarning)
            print("Filter cutoff has to be smaller than 0.5. Set it to 0.5")
            filter_cutoff = 0.5

        os.makedirs(filtered_image_folder, exist_ok=True)
        if img_num != -1 and total_num != -1:
            print(
                "Filtering ",
                img_path,
                "( Progress:",
                int(img_num / total_num * 100),
                "% )",
            )
        else:
            print("Filtering ", img_path)

        image_filtered = filter_single_image(
            img_path=img_path, filter_cutoff=filter_cutoff, resize_to_shape=resize_to_shape
        )
        if image_filtered is None:
            return None
        imagereader.image_write(filtered_img_tmp_file_path, image_filtered)
        del image_filtered
    return filtered_img_tmp_file_path


def filter_single_image_from_np_array(image, filter_cutoff, resize_to_shape=None, channel_specifiction=0):
    """
    Given an image as numpy array, this procedure resize it optionally to a target size and then applies
    an low-pass filter on the image.

    :param image: Image/tomogram as numpy array
    :param filter_cutoff: Absolute frequency cutoff
    :param resize_to_shape: (Optional) Resize the image after filtering
    :param channel_specification: 0 for channel first, -1 for channel last (default is 0)
    :return: Filtered numpy arraz
    """

    cutoff_factor = 1.0
    if resize_to_shape is not None:
        if not isinstance(resize_to_shape, (list, tuple)):
            'Scale the short side to 1024'
            #         h              w
            ar = image.shape[0] / image.shape[1]
            if len(image.shape)>2:
                if channel_specifiction == 0:
                    ar = image.shape[1] / image.shape[2]
            if ar < 1:
                height = resize_to_shape
                width = int(resize_to_shape / ar)
            else:
                height = int(resize_to_shape * ar)
                width = resize_to_shape
            resize_to_shape = [height, width]
    if len(image.shape) == 2:
        if channel_specifiction == 0:
            image = image[np.newaxis, ...]
        else:
            image = image[..., np.newaxis]



    # Do filtering
    mask = None
    image_filtered_result = None
    for i in range(image.shape[channel_specifiction]):
        if channel_specifiction == 0:
            img = image[i]
        else:
            img = image[:,:,i]

        if resize_to_shape is not None:
            cutoff_factor = img.shape[0] / resize_to_shape[0]
            from PIL import Image
            convert_to_array = Image.fromarray(img)
            resized = convert_to_array.resize((resize_to_shape[1], resize_to_shape[0]),
                                              resample=Image.BILINEAR)
            img = np.asarray(resized)
        if image_filtered_result is None:
            if channel_specifiction == 0:
                image_filtered_result = np.zeros(
                    shape=(image.shape[channel_specifiction],img.shape[0],img.shape[1]),
                )
            else:
                image_filtered_result = np.zeros(
                    shape=(img.shape[0], img.shape[1],image.shape[channel_specifiction]),
                )
        mask_size_0 = fft.next_fast_len(img.shape[0]) # next_power_of2(img.shape[0])
        mask_size_1 = fft.next_fast_len(img.shape[1]) # next_power_of2(img.shape[1] )
        filter_width = 2 * img.shape[1] * filter_cutoff * cutoff_factor
        if mask is None:
            mask = window(filter_width, (mask_size_0, mask_size_1))
        image_filtered = apply_fft_mask(img, mask)
        if channel_specifiction == 0:
            image_filtered_result[i] = image_filtered
        else:
            image_filtered_result[:, :, i] = image_filtered
        del img
        #image_filtered_result = np.squeeze(image_filtered_result)

    return image_filtered_result

def filter_single_image(img_path, filter_cutoff, resize_to_shape=None):
    """
    This procedure reads an abitary image, resize it optionally to a target size then apply
    an low-pass filter on the image.

    :param img_path: Path to the image/tomogram to be filtered
    :param filter_cutoff: Absolute frequency cutoff
    :param resize_to_shape: (Optional) Resize the image after filtering
    :return: Filtered and resized image/tomogram
    """
    try:
        image = imagereader.image_read(img_path)
    except Exception:
        print(img_path + " is corrupted. Ignore it.")
        return None
    return filter_single_image_from_np_array(image, filter_cutoff, resize_to_shape)




from cryolo.augmentation.AugmentationMethod import AugmentationMethod

import numpy as np

class ContrastNormalizationAugmentation(AugmentationMethod):

    def __init__(self, alpha_range : list):
        """
        :param alpha_range: Range for alpha. Alpha controls the normalization.
        """
        self.alpha_range = alpha_range

    def transform_image(self, image : np.array) -> np.array:
        """
        Spread or squeeze the pixel values.
        :param image: Input image
        :return: Modified image
        """
        rand_multiplyer = self.alpha_range[0] + np.random.rand() * (
                self.alpha_range[1] - self.alpha_range[0]
        )
        middle = np.median(image)
        np.subtract(image, middle, out=image)
        np.multiply(rand_multiplyer, image, out=image)
        np.add(middle, image, out=image)

        return image

    def transform_coords(self, object_coords : list, image_dims : list) -> list:
        return object_coords


import unittest
import os
import cryolo.preprocessing


class MyTestCase(unittest.TestCase):
    def test_is_helicon_with_particle_coords_should_be_false(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/Myo5ADP_0682.txt")
        is_helicon = cryolo.preprocessing.is_helicon_with_particle_coords(path)
        self.assertFalse(is_helicon)

    def test_is_eman1_helicion_should_be_true(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/Myo5ADP_0682.txt")
        is_eman1 = cryolo.preprocessing.is_eman1_helicion(path)
        self.assertTrue(is_eman1)



#
# COPYRIGHT
# All contributions by Thorsten Wagner:
# Copyright (c) 2017 - 2019, Thorsten Wagner.
# All rights reserved.
#
# ---------------------------------------------------------------------------
#         Do not reproduce or redistribute, in whole or in part.
#      Use of this code is permitted only under licence from Max Planck Society.
#            Contact us at thorsten.wagner@mpi-dortmund.mpg.de
# ---------------------------------------------------------------------------
#

from __future__ import print_function

import difflib
import os

import numpy as np

import cryolo.CoordsIO as CoordsIO
import cryolo.imagereader as imagereader
import cryolo.utils as utils

def get_image_size_distr(img_paths):
    sizes = []
    for timg in img_paths:
        size = imagereader.read_width_height(timg)
        if len(sizes) == 0:
            new = True
        else:
            new = True
            for known_size in sizes:
                if known_size[0] == size[0] and known_size[1] == size[1]:
                    known_size[2] = known_size[2] + 1
                    new = False
                    break
        if new:
            sizes.append([int(size[0]), int(size[1]), 1])
    sizes.sort(key=lambda x: x[2], reverse=True)
    return sizes

def find_image_annotation_pairs(annotations,images):
    img_names = list(map(os.path.basename, images))
    img_annot_pairs = []
    for ann in annotations:
        ann_without_ext = os.path.splitext(os.path.basename(ann))[0]
        cand_list = [name for name in img_names if ann_without_ext in name]
        try:
            cand_list_no_fileextension = list(map(os.path.basename, cand_list))
            corresponding_img_path = difflib.get_close_matches(
                ann_without_ext, cand_list_no_fileextension, n=1, cutoff=0
            )[0]
            corresponding_img_path = cand_list[
                cand_list_no_fileextension.index(corresponding_img_path)
            ]

        except IndexError:
            print("Cannot find corresponding image file for ", ann, " - Skipped.")
            continue
        index_image = img_names.index(corresponding_img_path)
        # Check if tomographyic data
        if imagereader.get_num_frames(images[index_image])>1:
            # For each unique z value generate one training tuple
            assert ann.endswith(".cbox"),"Only cbox files are supported for training on tomograms"
            boxes = CoordsIO.read_cbox_boxfile(ann)
            zvalues = CoordsIO.read_cbox_include_list(ann)
            #zvalues = np.unique([b.z for b in boxes])
            #print(zvalues,zvalues_)
            for z in zvalues:
                img_annot_pairs.append((images[index_image], ann, z))
        else:
            img_annot_pairs.append((images[index_image], ann))
    return img_annot_pairs

def find_image_annotation_pairs_by_dir(ann_dir, img_dir):
    if not os.path.exists(ann_dir):
        import sys

        print(
            "Your annotation folder does not exists: ",
            ann_dir,
            " Check your config file!",
        )
        sys.exit(1)

    if not os.path.exists(img_dir):
        import sys

        print(
            "Your image folder does not exists: ", img_dir, " Check your config file!"
        )
        sys.exit(1)


    img_files = []

    # Scan all image filenames
    for root, directories, filenames in os.walk(img_dir, followlinks=True):
        for filename in filenames:
            if filename.endswith(
                ("jpg", "png", "tiff", "tif", "mrc", "mrcs", "rec")
            ) and not filename.startswith("."):
                img_files.append(os.path.join(root, filename))

    # Read annotations

    annotations = []
    for root, directories, filenames in os.walk(ann_dir, followlinks=True):
        for ann in sorted(filenames):
            if ann.endswith((".box", ".txt", ".star", ".cbox")) and not ann.startswith("."):
                annotations.append(os.path.join(root,ann))
    img_annot_pairs = find_image_annotation_pairs(annotations,img_files)

    return img_annot_pairs



def parse_annotation(img_annot_pairs, grid_dims=None, anchor_size=None):

    all_imgs = []
    seen_labels = {}
    last_boxpath=None #Only for tomo data
    is_filament_data = False
    for train_tuble in img_annot_pairs:
        zpos=None
        if len(train_tuble)==3:
            filename, boxpath, zpos = train_tuble
        else:
            filename, boxpath = train_tuble

        img = {"object": []}
        img["filename"] = filename#img_files[index_image]
        img["boxpath"] = boxpath
        img["img_size"] = imagereader.read_width_height(img["filename"])
        if zpos is not None:
            img["z"] = zpos

        is_helicion_ptcl = is_helicon_with_particle_coords(boxpath)
        is_helicion_eman = is_eman1_helicion(boxpath)
        is_filament_cbox_file = is_filament_cbox(boxpath)
        if is_helicion_ptcl or is_helicion_eman or is_filament_cbox_file:
            grid_h = grid_dims[1]
            num_patches = grid_dims[2]
            cell_h = img["img_size"][1] / (num_patches * grid_h)

            if is_helicion_ptcl:
                filaments = CoordsIO.read_eman1_helicon(boxpath, int(cell_h))
            elif is_filament_cbox_file:
                filaments = CoordsIO.read_cbox_boxfile(boxpath, int(cell_h))
            else:
                filaments = CoordsIO.read_eman1_filament_start_end(
                    boxpath, int(cell_h)
                )
            for filament in filaments:
                fil_obj = []
                if filament is None or len(filament.boxes)==1:
                    continue
                for box in filament.boxes:
                    obj = {}
                    if zpos is not None and int(box.z) != int(zpos):
                        continue
                    box_xmin = int(box.x)
                    box_width = int(box.w)
                    box_height = int(box.h)
                    box_ymin = img["img_size"][1] - (int(box.y) + box_height)

                    box_xmax = box_xmin + box_width
                    box_ymax = box_ymin + box_height
                    obj["name"] = "particle"
                    obj["xmin"] = box_xmin
                    obj["ymin"] = box_ymin
                    obj["xmax"] = box_xmax
                    obj["ymax"] = box_ymax

                    fil_obj += [obj]

                #Set angles:
                for i in range(1,len(fil_obj)): # OLD len(fil_obj)

                    dy = fil_obj[i]["ymin"] - fil_obj[i-1]["ymin"]
                    dx = fil_obj[i]["xmin"] - fil_obj[i-1]["xmin"]
                    angle = utils.vector_to_direction(dy,dx)
                    fil_obj[i]["angle"] = angle

                if len(fil_obj)>0:
                    fil_obj[0]["angle"] = fil_obj[1]["angle"]
                    is_filament_data = True
                    img["object"] += fil_obj
                    if obj["name"] in seen_labels:
                        seen_labels[obj["name"]] += len(fil_obj)
                    else:
                        seen_labels[obj["name"]] = len(fil_obj)
        else:
            if os.stat(boxpath).st_size != 0:
                if boxpath.endswith((".star")):
                    box_width = int(anchor_size)
                    box_height = int(anchor_size)
                    boxes = CoordsIO.read_star_file(boxpath, (box_width + box_height) // 2)
                elif boxpath.endswith((".cbox")):
                    if last_boxpath is None or last_boxpath != boxpath:
                        boxes = CoordsIO.read_cbox_boxfile(boxpath)
                        last_boxpath = boxpath
                else:
                    boxes = CoordsIO.read_eman1_boxfile(boxpath)

                for box in boxes:

                    """
                    Box files are written with coordinate system with an origin in the top left corner. 
                    Each box file is specified by the lower left corner of the box and witdh and the hight of the box.
                    This has to be converted to a coordinate system with origin in the botten left corner.
                    """
                    if zpos is not None:
                        if int(box.z) != zpos:
                            continue
                    obj = {}
                    box_xmin = int(box.x)
                    box_width = int(box.w)
                    box_height = int(box.h)
                    box_ymin = img["img_size"][1] - (int(box.y) + box_height)

                    box_xmax = box_xmin + box_width
                    box_ymax = box_ymin + box_height
                    obj["name"] = "particle"
                    obj["xmin"] = box_xmin
                    obj["ymin"] = box_ymin
                    obj["xmax"] = box_xmax
                    obj["ymax"] = box_ymax
                    if zpos is not None:
                        obj["z"] = zpos

                    if zpos is None or box.z == zpos:
                        img["object"] += [obj]
                        if obj["name"] in seen_labels:
                            seen_labels[obj["name"]] += 1
                        else:
                            seen_labels[obj["name"]] = 1

        if len(img["object"]) >= 0:
            all_imgs += [img]
    result_dict = {}
    result_dict["images"] = all_imgs
    result_dict["labels"] = seen_labels
    result_dict["is_filament_data"] = is_filament_data
    return result_dict


def is_helicon_with_particle_coords(path):
    try:
        with open(path) as f:
            first_line = f.readline()
            f.close()
        return "#micrograph" in first_line
    except ValueError:
        return False

def is_filament_cbox(path):
    from pyStarDB import sp_pystardb as star
    try:
        starfile = star.StarFile(path)
        if 'cryolo' in starfile:
            is_filament = '_filamentid' in starfile['cryolo']
            return is_filament
    except Exception:
        return False

def is_eman1_helicion(path):
    try:
        if os.stat(path).st_size == 0:
            return False
        box_lines = np.atleast_2d(np.genfromtxt(path))
        return (
            len(box_lines) > 1
            and len(box_lines[0]) == 5
            and box_lines[0][4] == -1
            and box_lines[1][4] == -2
        )
    except ValueError:
        return False



import unittest
import cryolo.filament_tracer
from cryolo.utils import BoundBox
import numpy as np
import random
from cryolo.utils import Filament
import matplotlib.pyplot as plt
import os


class FilamentReplacerTest(unittest.TestCase):
    def test_is_valid(self):

        ref_angle = 0
        delta_angle = 10
        test_angle = 180

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertTrue(valid)

    def test_is_valid_2(self):
        ref_angle = 0
        delta_angle = 10
        test_angle = 0

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertTrue(valid)

    def test_is_valid_3(self):
        ref_angle = 0
        delta_angle = 10
        test_angle = 11

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertFalse(valid)

    def test_is_valid_4(self):
        ref_angle = 0
        delta_angle = 10
        test_angle = 169

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertFalse(valid)

    def test_is_valid_5(self):
        ref_angle = 70
        delta_angle = 10
        test_angle = 81

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertFalse(valid)

    def test_is_valid_6(self):
        ref_angle = 180
        delta_angle = 10
        test_angle = 5

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertTrue(valid)

    def test_is_valid_7(self):

        ref_angle = 178
        delta_angle = 10
        test_angle = 2

        valid = cryolo.filament_tracer.is_valid(ref_angle, delta_angle, test_angle)

        self.assertTrue(valid)

    def test_angle_between_two_boxes_90degree(self):
        boxa = BoundBox(x=0, y=1, w=10, h=10)
        boxb = BoundBox(x=1, y=0, w=10, h=10)

        angle = cryolo.filament_tracer.angle_between_two_boxes(boxa, boxb)

        self.assertAlmostEqual(first=90, second=angle, places=1)

    def test_angle_between_two_boxes_0degree(self):
        boxa = BoundBox(x=3, y=3, w=10, h=10)
        boxb = BoundBox(x=3, y=3, w=10, h=10)

        angle = cryolo.filament_tracer.angle_between_two_boxes(boxa, boxb)

        self.assertAlmostEqual(first=0, second=angle, places=1)

    def test_angle_between_two_boxes_45degree(self):
        boxa = BoundBox(x=3, y=3, w=10, h=10)
        boxb = BoundBox(x=0, y=3, w=10, h=10)

        angle = cryolo.filament_tracer.angle_between_two_boxes(boxa, boxb)

        self.assertAlmostEqual(first=45, second=angle, places=1)

    def test_angle_between_two_boxes_180degree(self):
        boxa = BoundBox(x=1, y=1, w=10, h=10)
        boxb = BoundBox(x=-1, y=-1, w=10, h=10)

        angle = cryolo.filament_tracer.angle_between_two_boxes(boxa, boxb)

        self.assertAlmostEqual(first=180, second=angle, places=1)

    def test_next_valid_box(self):

        angle_image = np.zeros(shape=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))
        for i in range(250, 750, 5):
            boxes.append(BoundBox(x=512 - 10 - width / 2, y=i, w=width / 2, h=width))

        assigned = []
        assigned.append(boxes[0])

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, search_radius=60, angle_delta=10
        )
        candidate_boxes = [box for box in boxes if box not in assigned]
        next_box_index = tracer.next_valid_box(
            filament=None, candidate_boxes=candidate_boxes, ref_box=boxes[0]
        )

        self.assertEqual(boxes[1], candidate_boxes[0])

    def test_correct_merging(self):

        angle_image = np.zeros(shape=(1024, 1024))
        angle_image[250:750, 512 - 40 : 512 + 40] = 90

        angle_image[750:1000, 512 - 40 : 512 + 40] = 90

        boxes = []
        width = 30
        for i in range(250, 300, 10):
            box = BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width)
            boxes.append(box)
            print(box.x, box.y)
        print("First part:", len(boxes))
        len1 = len(boxes)
        for i in range(300, 350, 10):
            box = BoundBox(
                x=512 - 15 - width / 2, y=i - width / 2, w=width / 2, h=width
            )
            boxes.append(box)
            print(box.x, box.y)
        print("Secondpart part:", len(boxes) - len1)
        len1 = len(boxes)
        assigned = []
        assigned.append(boxes[0])

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, search_radius=30, angle_delta=10
        )
        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments))

    def test_correct_merging_numboxes(self):

        angle_image = np.zeros(shape=(1024, 1024))
        angle_image[250:750, 512 - 40 : 512 + 40] = 90

        angle_image[750:1000, 512 - 40 : 512 + 40] = 90

        boxes = []
        width = 30
        for i in range(250, 300, 10):
            box = BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width)
            boxes.append(box)
            print(box.x, box.y)
        print("First part:", len(boxes))
        len1 = len(boxes)
        for i in range(290, 340, 10):
            box = BoundBox(
                x=512 - 16 - width / 2, y=i - width / 2, w=width / 2, h=width
            )
            boxes.append(box)
            print(box.x, box.y)
        print("Secondpart part:", len(boxes) - len1)
        len1 = len(boxes)
        assigned = []
        assigned.append(boxes[0])

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, search_radius=60, angle_delta=10
        )
        filaments = tracer.trace_filaments()

        self.assertEqual(5, len(filaments[0].boxes))
        self.assertEqual(5, len(filaments[1].boxes))

    def test_split_filament(self):

        boxes = []
        width = 30
        boxes.append(BoundBox(x=100 - width / 2, y=100 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=101 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=102 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=103 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=104 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=105 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=106 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=107 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=108 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=100 - width / 2, y=109 - width / 2, w=width, h=width))

        boxes.append(BoundBox(x=101 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=102 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=103 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=104 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=105 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=106 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=107 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=108 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=109 - width / 2, y=109 - width / 2, w=width, h=width))
        boxes.append(BoundBox(x=110 - width / 2, y=109 - width / 2, w=width, h=width))

        fil = Filament(boxes)
        splitted_filaments = cryolo.filament_tracer.split_filament_by_straightness_rec(
            [fil],
            straightness_method=cryolo.filament_tracer._get_straightness,
            straightness_threshold=0.9,
        )

        self.assertEqual(len(splitted_filaments), 2)

    def test_split_filaments_2(self):
        import cryolo.CoordsIO

        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/Myo5a_ELC_Rigor_00105_wrong_splitting.box",
        )
        boxes = cryolo.CoordsIO.read_eman1_boxfile(path)
        filament = Filament()
        filament.boxes = boxes
        segments = cryolo.filament_tracer.split_filament_by_straightness(filament)

        num_boxes = 0
        for seg in segments:
            num_boxes += len(seg.boxes)
        self.assertEqual(75, num_boxes)

    def test_filament_direction(self):
        boxes = []
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512, y=i, w=30, h=30))

        filament = Filament()
        filament.boxes = boxes

        fil_dir = cryolo.filament_tracer.filament_direction(filament)

    def test_replace_boxes_single_line(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(1, len(filaments))
        self.assertTrue(
            is_ordered(y_to_list(filaments[0])), "Filament 0 is not ordered"
        )

    def test_replace_boxes_single_line_shuffeld_length_test(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:750, (512 - 8) : (512 + 8)] = 90

        boxes = []
        width = 30
        for i in range(300, 500, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))
        random.Random(5).shuffle(boxes)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(1, len(filaments), "Only one filament should be detected.")
        self.assertEqual(
            len(filaments[0].boxes),
            len(boxes),
            "Input and out have not the same length.",
        )

    def test_replace_boxes_single_line_shuffeld_num_boxes(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 450, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))
        random.Random(5).shuffle(boxes)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(
            len(filaments[0].boxes),
            len(boxes),
            "Input and out have not the same length.",
        )

    def test_replace_boxes_single_line_shuffeld_ordered(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 450, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))
        random.Random(5).shuffle(boxes)
        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image
        )

        filaments = tracer.trace_filaments()

        for i in range(len(filaments)):
            self.assertTrue(
                is_ordered(y_to_list(filaments[i])),
                "Filament " + str(i) + " is not ordered",
            )

    def test_replace_boxes_two_lines(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:400, 512 - 8 : 512 + 8] = 90
        angle_image[500:750, 512 - 8 : 512 + 8] = 90
        boxes = []
        width = 30
        for index, i in enumerate(range(250, 750, 8)):
            box = BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width)
            print(
                index,
                int(box.x),
                int(box.y),
                angle_image[int(box.y + box.h / 2), int(box.x + box.w / 2)],
            )
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, angle_delta=10
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments))
        self.assertTrue(
            is_ordered(y_to_list(filaments[0])), "Filament 0 is not ordered"
        )
        self.assertTrue(
            is_ordered(y_to_list(filaments[1])), "Filament 1 is not ordered"
        )

    def test_replace_boxes_real_example(self):
        path = os.path.join(
            os.path.dirname(__file__),
            "../resources/Actin-ADP-BeFx_0250_60e_DW_downscaled_enhanced.txt",
        )
        angle_image = np.loadtxt(path)
        angle_image = np.flipud(angle_image)

        path = os.path.join(
            os.path.dirname(__file__), "../resources/Actin-ADP-BeFx_0250_60e_DW.box"
        )
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])  # 4096-int(row[1])-box_height
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)
        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=60,
            angle_delta=10,
            rescale_factor=0.25,
            box_distance=5,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments))

    def test_replace_boxes_real_example_2(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/enhanced.txt")
        angle_image = np.loadtxt(path)
        # plt.imshow(angle_image)
        angle_image = np.flipud(angle_image)

        # plt.show()
        path = os.path.join(
            os.path.dirname(__file__), "../resources/Actin-ADP-BeFx_0254_60e_DW.box"
        )
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])  # 4096-int(row[1])-box_height
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=60,
            angle_delta=10,
            rescale_factor=0.25,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments))

    def test_replace_boxes_real_example_3(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/enhanced_3251.txt")
        angle_image = np.loadtxt(path)
        # plt.imshow(angle_image)
        angle_image = np.flipud(angle_image)

        # plt.show()
        path = os.path.join(os.path.dirname(__file__), "../resources/gammaADP-3251.box")
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])  # 4096-int(row[1])-box_height
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=96,
            angle_delta=10,
            rescale_factor=0.25,
        )

        filaments = tracer.trace_filaments()

        print(len(filaments[0].boxes))

        self.assertEqual(1, len(filaments))

    def test_replace_boxes_real_example_6(self):
        # Example contains only one filament. Merging should do the job.
        # However, it failed in the past
        path = os.path.join(os.path.dirname(__file__), "../resources/01.txt")
        angle_image = np.loadtxt(path)
        angle_image = np.flipud(angle_image)
        print("Shape:", angle_image.shape)
        # plt.imshow(angle_image)

        path = os.path.join(os.path.dirname(__file__), "../resources/01.box")
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        xs = []
        ys = []
        rescale_factor = 0.2668
        for row in box_lines:
            xs.append(int((row[0] + row[2] / 2) * rescale_factor))
            ys.append(int((row[1] + row[3] / 2) * rescale_factor))
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        # plt.scatter(xs, ys, s=100, edgecolors="r", facecolors='none')
        # plt.show()
        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=53.361,
            angle_delta=10,
            rescale_factor=rescale_factor,
            box_distance=5,
            rescale_factor_x=0.2760,
            rescale_factor_y=0.2668,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(1, len(filaments))

    def test_distance_perpendicluar_to_box_direction(self):
        box1 = BoundBox(x=5, y=0, w=30, h=30)
        box2 = BoundBox(x=10, y=1, w=30, h=30)

        dist = cryolo.filament_tracer.distance_perpendicluar_to_box_direction(
            ref_box=box1, candiate_box=box2, ref_angle=0
        )

        self.assertAlmostEqual(first=1, second=dist, places=4)

    def test_trace_filaments_real_example_4(self):
        path = os.path.join(os.path.dirname(__file__), "../resources/3200_enhanced.txt")
        angle_image = np.loadtxt(path)
        angle_image = np.flipud(angle_image)
        path = os.path.join(os.path.dirname(__file__), "../resources/gammaADP-3200.box")
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])  # 4096-int(row[1])-box_height
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=256 * 1.5 * 0.25,
            angle_delta=10,
            box_distance=10,
            rescale_factor=0.25,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(1, len(filaments))

    def test_trace_filaments_real_example_5(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/greg_angle_image.txt"
        )
        angle_image = np.loadtxt(path)
        # plt.imshow(angle_image)
        angle_image = np.flipud(angle_image)

        # plt.show()
        path = os.path.join(os.path.dirname(__file__), "../resources/greg_mrc_01.box")
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])  # 4096-int(row[1])-box_height
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=55,
            angle_delta=10,
            rescale_factor=0.2668,
            min_number_boxes=6,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments))

    def test_trace_filaments_real_example_7_pickonefilament(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/Myo5a_ELC_ADP03334_enhanced.txt"
        )
        angle_image = np.loadtxt(path)
        # plt.imshow(angle_image)
        angle_image = np.flipud(angle_image)

        # plt.show()
        path = os.path.join(
            os.path.dirname(__file__), "../resources/Myo5a_ELC_ADP03334.box"
        )
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=68.30,
            angle_delta=10,
            rescale_factor=0.2668,
            rescale_factor_x=0.2760,
            rescale_factor_y=0.2668,
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(1, len(filaments))

    @unittest.skip("Need to rewrite this this")
    def test_trace_filaments_real_example_8_pickonefilament(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/Myo5a_ELC_ADP03719.txt"
        )
        angle_image = np.loadtxt(path)
        # plt.imshow(angle_image)
        angle_image = np.flipud(angle_image)

        # plt.show()
        path = os.path.join(
            os.path.dirname(__file__), "../resources/Myo5a_ELC_ADP03719.box"
        )
        box_lines = np.atleast_2d(np.genfromtxt(path))
        boxes = []
        for row in box_lines:
            box_xmin = int(row[0])
            box_width = int(row[2])
            box_height = int(row[3])
            box_ymin = int(row[1])
            box = BoundBox(x=box_xmin, y=box_ymin, w=box_width, h=box_height)
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes,
            orientation_image=angle_image,
            search_radius=68.30,
            angle_delta=10,
            rescale_factor=0.2668,
            rescale_factor_x=0.2760,
            rescale_factor_y=0.2668,
        )

        filaments = tracer.trace_filaments()
        for f_i, f in enumerate(filaments):
            print("FI:", f_i)
            for b in f.boxes:
                print("x", b.x, "y", b.y)

        import cryolo.utils as utils

        minimum_number_boxes = 6
        box_distance = 26
        resamples_filaments = utils.resample_filaments(filaments, box_distance)

        # Make straight filaments

        resamples_filaments = cryolo.filament_tracer.split_filaments_by_straightness(
            resamples_filaments,
            straightness_method=cryolo.filament_tracer._get_straightness,
            straightness_threshold=0.95,
        )

        # Min number of boxes filter:
        resamples_filaments = cryolo.filament_tracer.filter_filaments_by_num_boxes(
            resamples_filaments, minimum_number_boxes
        )

        for f_i, f in enumerate(resamples_filaments):
            print("F:", f_i)
            for b in f.boxes:
                print("x", b.x, "y", b.y)

        self.assertEqual(1, len(resamples_filaments))

    def test_replace_boxes_two_lines_shuffeld(self):

        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:400, 512 - 8 : 512 + 8] = 90
        angle_image[500:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 750, 8):
            boxes.append(BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width))
        random.Random(5).shuffle(boxes)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, angle_delta=10, search_radius=30
        )

        filaments = tracer.trace_filaments()

        self.assertEqual(2, len(filaments), "Containes more elemetns as expected")
        self.assertTrue(
            is_ordered(y_to_list(filaments[0])), "Filament 0 is not ordered"
        )
        self.assertTrue(
            is_ordered(y_to_list(filaments[1])), "Filament 1 is not ordered"
        )

    def test_merge_filaments(self):
        angle_image = np.zeros(shape=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        for i in range(250, 750, 8):
            box = BoundBox(x=512 - width / 2, y=i - width / 2, w=width, h=width)
            # box.info = i
            boxes.append(box)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, search_radius=20, angle_delta=10
        )
        filaments = tracer.trace_filaments()
        print("Fil length:", len(filaments[0].boxes))
        fil_a = Filament()
        fil_a.boxes = filaments[0].boxes[: len(filaments[0].boxes) // 2]
        fil_b = Filament()
        fil_b.boxes = filaments[0].boxes[len(filaments[0].boxes) // 2 :]

        new_filament = tracer.merge_filament(filament_a=fil_a, filament_b=fil_b)

        self.assertIsNotNone(new_filament)

    def test_in_search_region_expected_true(self):

        ref_box = BoundBox(x=1, y=1, w=30, h=30)
        candidate_box = BoundBox(x=2, y=2, w=30, h=30)
        ref_angle = 45
        angle_delta = 10

        result = cryolo.filament_tracer.in_angular_search_region(
            ref_box, candidate_box, ref_angle, angle_delta
        )

        self.assertTrue(result)

    def test_in_search_region_mirrored_expected_true(self):

        ref_box = BoundBox(x=1, y=1, w=30, h=30)
        candidate_box = BoundBox(x=-2, y=-2, w=30, h=30)
        ref_angle = 45
        angle_delta = 10

        result = cryolo.filament_tracer.in_angular_search_region(
            ref_box, candidate_box, ref_angle, angle_delta
        )

        self.assertTrue(result)

    def test_in_search_region_expected_false(self):

        ref_box = BoundBox(x=0, y=1, w=30, h=30)
        candidate_box = BoundBox(x=1, y=0, w=30, h=30)
        ref_angle = 0
        angle_delta = 10

        result = cryolo.filament_tracer.in_angular_search_region(
            ref_box, candidate_box, ref_angle, angle_delta
        )

        self.assertFalse(result)

    def test_merge_filement_manyboxes_to_one(self):
        angle_image = np.random.randint(low=0, high=60 + 1, size=(1024, 1024))
        angle_image[250:750, 512 - 8 : 512 + 8] = 90

        boxes = []
        width = 30
        """
        Generates the coordinates:
        497 339
        497 347
        497 355
        497 363
        497 371
        497 379
        497 387
        497 395
        497 403
        497 411
        497 419
        """
        for i in range(354, 435, 8):
            x = 512 - width / 2
            boxes.append(BoundBox(x=x, y=i - width / 2, w=width, h=width))
        fila = Filament(boxes)

        boxes = []
        boxes.append(BoundBox(x=497, y=427, w=width, h=width))
        filb = Filament(boxes)

        tracer = cryolo.filament_tracer.FilamentTracer(
            boxes, orientation_image=angle_image, search_radius=20, angle_delta=10
        )
        new_filament = tracer.merge_filament(filament_a=fila, filament_b=filb)
        self.assertIsNotNone(new_filament)

    def test_nms_imag72_twofilaments(self):
        import cryolo.CoordsIO as coords

        path = os.path.join(
            os.path.dirname(__file__), "../resources/ActinLifeAct_00072_nms_bug.box"
        )
        filaments = coords.read_eman1_helicon(path)
        filaments_nms = cryolo.filament_tracer.nms_for_filaments(
            filaments, iou_thresh=0.5
        )

        self.assertEqual(len(filaments_nms), 1)

    def test_nms_for_filaments_three_overlapping_2(self):
        width = 10
        boxes = []
        boxes2 = []
        boxes3 = []

        # Generating this:
        #          ----------------------
        # ---------
        #       -------
        # Extepecting:
        #          ----------------------
        # ---------
        for i in range(100, 300, 10):
            boxes.append(BoundBox(x=i, y=50, w=width, h=width))
        for i in range(0, 100, 10):
            boxes2.append(BoundBox(x=i, y=50, w=width, h=width))
        for i in range(70, 150, 10):
            boxes3.append(BoundBox(x=i, y=50, w=width, h=width))

        fila = Filament(boxes)
        filb = Filament(boxes2)
        filc = Filament(boxes3)

        filaments = cryolo.filament_tracer.nms_for_filaments([filc, filb, fila])
        filaments_nms = cryolo.filament_tracer.nms_for_filaments(
            filaments, iou_thresh=0.5
        )

        self.assertEqual(2, len(filaments_nms))

    def test_nms_for_filaments_two_in_three(self):
        """
        Generating two filaments that cross at one box. They should be splitted into 3 filaments
        """
        # nms_for_filaments
        width = 10
        boxes = []
        boxes2 = []
        for i in range(0, 200, 10):
            boxes.append(BoundBox(x=i, y=50, w=width, h=width))
            boxes2.append(BoundBox(x=100, y=i, w=width, h=width))
        fila = Filament(boxes)
        filb = Filament(boxes2)

        filaments = cryolo.filament_tracer.nms_for_filaments([fila, filb])
        len_total = 0
        for fil in filaments:
            len_total += len(fil.boxes)
        """
        Expected result: The two filaments are splitted in into three filaments
        One hsould have length 20, on should have length 5 and one length 14. One overlapping
        box of on filaments (the shorter in general) is deleted.
        """
        self.assertEqual(
            len(filaments),
            3,
            "The number of splitted filaments should be 3 but is "
            + str(len(filaments)),
        )
        self.assertEqual(
            len_total,
            20 + 14 + 5,
            "The total lenght should be 39 but is " + str(len_total),
        )

    def test_nms_for_filaments_complete_overlap(self):
        """
        Generating two filaments that cross at one box. They should be splitted into 3 filaments
        """
        # nms_for_filaments
        width = 10
        boxes = []
        boxes2 = []
        for i in range(0, 200, 10):
            boxes.append(BoundBox(x=i, y=50, w=width, h=width))
        for i in range(0, 100, 10):
            boxes2.append(BoundBox(x=i, y=50, w=width, h=width))
        fila = Filament(boxes)
        filb = Filament(boxes2)

        filaments = cryolo.filament_tracer.nms_for_filaments([fila, filb])

        """
        Expected result: Only one filament with length 20
        """
        self.assertEqual(
            len(filaments),
            1,
            "The number of splitted filaments should be 1 but is "
            + str(len(filaments)),
        )
        self.assertEqual(
            len(filaments[0].boxes),
            20,
            "The total lenght should be 20 but is " + str(len(filaments[0].boxes)),
        )


def x_to_list(filament):
    xlist = []
    for box in filament.boxes:
        xlist.append(box.x)
    return xlist


def y_to_list(filament):
    ylist = []
    for box in filament.boxes:
        ylist.append(box.y)
    return ylist


def is_ordered(alist):
    descending = True
    for i in range(1, len(alist)):
        if alist[i] > alist[i - 1]:
            descending = False
            break
    ascending = True
    for i in range(len(alist) - 1):
        if alist[i] > alist[i + 1]:
            ascending = False
            break
    return descending or ascending


if __name__ == "__main__":
    unittest.main()



import keras.backend as K
import numpy as np
from keras.callbacks import Callback


class SGDRScheduler(Callback):
    """Cosine annealing learning rate scheduler with periodic restarts.
    # Usage
        ```python
            schedule = SGDRScheduler(min_lr=1e-5,
                                     max_lr=1e-2,
                                     steps_per_epoch=np.ceil(epoch_size/batch_size),
                                     lr_decay=0.9,
                                     cycle_length=5,
                                     mult_factor=1.5)
            model.fit(X_train, Y_train, epochs=100, callbacks=[schedule])
        ```
    # Arguments
        min_lr: The lower bound of the learning rate range for the experiment.
        max_lr: The upper bound of the learning rate range for the experiment.
        steps_per_epoch: Number of mini-batches in the dataset. Calculated as `np.ceil(epoch_size/batch_size)`.
        lr_decay: Reduce the max_lr after the completion of each cycle.
                  Ex. To reduce the max_lr by 20% after each cycle, set this value to 0.8.
        cycle_length: Initial number of epochs in a cycle.
        mult_factor: Scale epochs_to_restart after each full cycle completion.
    # References
        Blog post: jeremyjordan.me/nn-learning-rate
        Original paper: http://arxiv.org/abs/1608.03983
    """

    def __init__(
        self,
        min_lr,
        max_lr,
        steps_per_epoch,
        lr_decay=1,
        cycle_length=10,
        mult_factor=2,
    ):

        self.min_lr = min_lr
        self.max_lr = max_lr
        self.lr_decay = lr_decay

        self.batch_since_restart = 0
        self.next_restart = cycle_length

        self.steps_per_epoch = steps_per_epoch

        self.cycle_length = cycle_length
        self.mult_factor = mult_factor

        self.history = {}

    def clr(self):
        """Calculate the learning rate."""
        fraction_to_restart = self.batch_since_restart / (
            self.steps_per_epoch * self.cycle_length
        )
        lr = self.min_lr + 0.5 * (self.max_lr - self.min_lr) * (
            1 + np.cos(fraction_to_restart * np.pi)
        )
        return lr

    def on_train_begin(self, logs={}):
        """Initialize the learning rate to the minimum value at the start of training."""
        logs = logs or {}
        K.set_value(self.model.optimizer.lr, self.max_lr)

    def on_batch_end(self, batch, logs={}):
        """Record previous batch statistics and update the learning rate."""
        logs = logs or {}
        self.history.setdefault("lr", []).append(K.get_value(self.model.optimizer.lr))
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        self.batch_since_restart += 1
        K.set_value(self.model.optimizer.lr, self.clr())

    def on_epoch_end(self, epoch, logs={}):
        """Check for end of current cycle, apply restarts when necessary."""
        if epoch + 1 == self.next_restart:
            self.batch_since_restart = 0
            self.cycle_length = np.ceil(self.cycle_length * self.mult_factor)
            self.next_restart += self.cycle_length
            self.max_lr *= self.lr_decay
            self.best_weights = self.model.get_weights()

    def on_train_end(self, logs={}):
        """Set weights to the values from the end of the most recent cycle for best performance."""
        if self.best_weights:
            self.model.set_weights(self.best_weights)



import unittest
from cryolo import eval
import os


class MyTestCase(unittest.TestCase):
    def test_readbox(self):
        path = os.path.join(
            os.path.dirname(__file__), "../resources/2018_08_14_BBSXL_AuXPos1_02791.box"
        )
        boxes = eval.read_box(path, 0)

        self.assertEqual(len(boxes), 1)


if __name__ == "__main__":
    unittest.main()



from .metric import Metric

from dataclasses import dataclass
from .boxdataprovider import BoxDataProvider, ImageBoxData
from .pathsboxdataprovider import PathsDataProvider
from cryolo.evaluation.pathsfilamentprovider import PathsFilamentProvider
from .recallprecisionmetric3d import RecallPrecisionMetric3D
from cryolo import CoordsIO
from typing import List, Callable, Tuple

import argparse
import numpy as np
import os
import glob


@dataclass
class EvaluationResult:
    metric : str
    value : float
    number_datapoints : int
    dataset : str
    threshold : float = None

from cryolo.evaluation.resultsview import SimpleView, ResultsView

def make_imageboxdata_pairs(
    seta: List[ImageBoxData],
    setb: List[ImageBoxData],
    identifier_matching_method: Callable[[str, str], bool],
) -> List[Tuple[ImageBoxData, ImageBoxData]]:
    pairs = []
    index_a_found = []
    index_b_found = []
    for index_a, boxdataA in enumerate(seta):
        for index_b, boxdataB in enumerate(setb):

            if identifier_matching_method(boxdataA.indentifer, boxdataB.indentifer):
                if index_a not in index_a_found and index_b not in index_b_found:

                    pairs.append((boxdataA, boxdataB))
                    index_a_found.append(index_a)
                    index_b_found.append(index_b)

    for index_a, boxdataA in enumerate(seta):
        if index_a not in index_a_found:
            pairs.append((boxdataA, ImageBoxData(None,None)))
    for index_b, boxdataB in enumerate(setb):
        if index_b not in index_b_found:
            pairs.append((ImageBoxData(None,None), boxdataB))

    return pairs


def group_evaluation_results(results : List[EvaluationResult]) -> List[EvaluationResult]:
    groups = set([e.dataset for e in results])
    group_result = []
    for group in groups:
        results_same_group = [e for e in results if e.dataset == group]
        metrics = set([e.metric for e in results_same_group])
        for metric in metrics:
            values = [e.value for e in results_same_group if e.metric == metric]
            res = EvaluationResult(
                metric=metric,
                value=np.mean(values),
                number_datapoints=len(values),
                dataset=group
            )
            group_result.append(res)

    return group_result

def evaluate(
    gtProvider: BoxDataProvider,
    measuredProvider: BoxDataProvider,
    metrics: List[Metric],
    identifier_matching_method=BoxDataProvider.path_matching_filename_eq,
) -> List[EvaluationResult]:
    """
    :return: Dictionary of averages and dictionary of number of datapoints.
    """

    imageboxdataGT = gtProvider.get_boxes()
    imageboxdataM = measuredProvider.get_boxes()

    pairs = make_imageboxdata_pairs(imageboxdataGT, imageboxdataM, identifier_matching_method=identifier_matching_method)

    # Calculate all evaluation results
    eval_results = []
    for gtData, mData in pairs:
        for metric in metrics:
            evaluation_result = metric.eval(mData.data, gtData.data)
            for metricResult in evaluation_result:
                eres = EvaluationResult(metric=metricResult,
                                 value=evaluation_result[metricResult],
                                 dataset=mData.group,
                                 number_datapoints=1) # 1 because it is only one image.
                eval_results.append(eres)

    return group_evaluation_results(eval_results)

def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Evaluation script for Tomography")

    parser.add_argument(
        "-gtb",
        "--groundtruth",
        type=str,
        required=True,
        help="Path box files with ground truth box files",
    )

    parser.add_argument(
        "-pb",
        "--predicted",
        type=str,
        required=True,
        help="Path box files with ground truth cbox files",
    )

    parser.add_argument(
        "--filament",
        action="store_true",
        help="Use this flag if your data is filament data",
    )

    parser.add_argument(
        "-bd",
        "--boxdistance",
        type=int,
        help="(FIlAMENTS ONLY) Box distance when resampling should be used.",
    )

    return parser


def run(args : List[str], metrics : List[Metric], view : ResultsView) -> List[EvaluationResult]:

    # Get path path to ground truth boxes and measured boxes
    mask_gt = "*.cbox"
    path_gt = args.groundtruth

    mask_measured = "*.cbox"
    path_measured = args.predicted

    is_filament = args.filament
    filament_box_distance = args.boxdistance

    # Get relevant files
    measured_boxfile_paths = glob.glob(os.path.join(path_measured, mask_measured))
    gt_boxfile_paths = glob.glob(os.path.join(path_gt, mask_gt))

    # Find file pairs
    if is_filament:
        mProvider = PathsFilamentProvider(
            paths=measured_boxfile_paths,
            reading_method=CoordsIO.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
            resampling_distance=filament_box_distance,
        )

        gtProvider = PathsFilamentProvider(
            paths=gt_boxfile_paths,
            reading_method=CoordsIO.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
            resampling_distance=filament_box_distance,
        )
    else:
        mProvider = PathsDataProvider(
            paths=measured_boxfile_paths,
            reading_method=CoordsIO.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )

        gtProvider = PathsDataProvider(
            paths=gt_boxfile_paths,
            reading_method=CoordsIO.read_cbox_boxfile,
            group_method=BoxDataProvider.group_method_lastfolder,
        )


    # Evaluate all metric on the box pairs
    results = evaluate(gtProvider, mProvider, metrics)

    # Print statistics ( ResultsView )
    view.run(results)

    return results

def _main_():

    parser = create_parser()
    args = parser.parse_args()

    view = SimpleView()

    metrics = []
    metrics.append(RecallPrecisionMetric3D(iou_thresh=0.5))

    run(args, view=view, metrics=metrics)




if __name__ == "__main__":
    _main_()
